#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include "isvd.h"

double ComputeResidual(int N, double* A, double* S, double* U, double* V);
extern "C" void dgemm_(char* transa, char* transb, int* m, int* n, int* k, double* alpha, double* A, int* lda, double* B, int* ldb, double* beta, double* C, int* ldc);

int main(){
	const int N = 1000; // Matrix size
	const int lb = 100; // Length of the band
	const int P = 4; // Number of threads
	const int count_to_show = 5;
	double *A, *values, *U, *V, *work, *AA;
	int i, j;
	
	A = (double*)malloc(sizeof(double) * N * N);
	values = (double*)malloc(sizeof(double) * N);
	U = (double*)malloc(sizeof(double) * N * N);
	V = (double*)malloc(sizeof(double) * N * N);
	work = (double*)malloc(sizeof(double) * GetWorkSize(N, lb, P, 1));
	AA = (double*)malloc(sizeof(double) * N * N);
	
	// Set the matrix
	srand(0);
	for(i = 0; i < N * N; i++){
		A[i] = (double)rand() / (double)RAND_MAX - 0.5;
	}
	// Save the matrix data (in DGESLV, the matrix data will be destroyed)
	memcpy(AA, A, sizeof(double) * N * N);

	// Singular value decomposition with I-SVD
	DGESLV(N, lb, A, values, U, V, work, P, 0);

	// Show the answer partly
	printf("Singular values = \n");
	for(i = 0; i < count_to_show; i++){
		printf("%le\n", values[i]);
	}
	printf("Matrix U = \n");
	for(i = 0; i < count_to_show; i++){
		for(j = 0; j < count_to_show; j++){
			printf("%le, ", U[j * N + i]);
		}
		printf("\n");
	}
	printf("Matrix V = \n");
	for(i = 0; i < count_to_show; i++){
		for(j = 0; j < count_to_show; j++){
			printf("%le, ", V[j * N + i]);
		}
		printf("\n");
	}
	
	printf("||A - USV^T||_F = %e\n", ComputeResidual(N, AA, values, U, V));
	
	free(A); free(values); free(U); free(V); free(work); free(AA);
}

double ComputeResidual(int N, double* A, double* S, double* U, double* V){
	// Compute the residual A - USV^T
	int i, j;
	double res = 0.0, one = 1.0, mone = -1.0;
	
	for(j = 0; j < N; j++){
		for(i = 0; i < N; i++){
			U[j * N + i] *= S[j];
		}
	}
	
	dgemm_("N", "T", &N, &N, &N, &one, U, &N, V, &N, &mone, A, &N);
	
	for(j = 0; j < N; j++){
		for(i = 0; i < N; i++){
			res += A[j * N + i] * A[j * N + i];
		}
	}
	
	return sqrt(res);
}
